"""This module contains the CharacterCipher class."""

from cStringIO import StringIO


class CharacterCipher:

    """This is a useful baseclass for character-oriented ciphers.

    Subclasses of this class do not need to implement encrypt or decrypt.
    Instead, they must implement encryptCharacter and decryptCharacter.

    """

    def encrypt(self, s):
        """Encrypt and return s."""
        return self._map(self.encryptCharacter, s)

    def decrypt(self, s):
        """Decrypt and return s."""
        return self._map(self.decryptCharacter, s)

    def _map(self, f, s):
        """Apply f to each character in s and return the new s.

        Use a cStringIO buffer.
        
        """
        buf = StringIO()
        for c in s:
            buf.write(f(c))
        return buf.getvalue()


# The following is for testing.
class _Test(CharacterCipher):
    def encryptCharacter(self, c): return 'z'
    def decryptCharacter(self, c): return 'a'


# Do some testing.
if __name__ == '__main__':
    cipher = _Test()
    assert cipher.encrypt("foo") == "zzz"
    assert cipher.decrypt("foo") == "aaa"



"""This file contains the MainGUI class."""

import gtk
import gnome
import gnome.ui
from os.path import join

import Const
from LibGladeApplication import LibGladeApplication
from GtkAttributesFacade import GtkAttributesFacade 
from ciphergui.CipherGUIList import cipherGUIList


class MainGUI(LibGladeApplication):

    """This is the main window for the application.
    
    HACK:  To do cut, copy, and paste I'm using the emit method.  I suspect
    that the PyGTK API will eventually update the TextView class to include the
    methods *_clipboard just as the C API does.
    
    The following private variables are used:

    _last_text_view - This is the last TextView that was focused.

    _cipherGUI - This is an instance of a class from the ciphergui package
    associated with the currently selected cipher.

    _cipherGUIInstances - Whenever a new ciphergui instance is created, it is
    cached here so that its key is not lost when the user switches between
    ciphers.

    """

    def __init__(self):
        """Init the GUI."""
        LibGladeApplication.__init__(self, 
            join(Const.GLADEDIR, "gcipher.glade"))
        self._last_text_view = self.decrypted_text_view
        self._cipherGUIInstances = {}
        self._initCipherMenu()

    def _initCipherMenu(self):
        """Initialize the Cipher menu.

        This menu must be dynamic since it must contain the list of ciphers, so
        it can't be created just using Glade.  Place the menu items before the 
        separator and the "Key..." menu item.

        """
        group = None
        l = cipherGUIList[:]
        l.reverse()
        for i in l:
            (className, guiName, labelName, description) = i
            menuItem = gtk.RadioMenuItem(group, labelName)
            if not group:
                group = menuItem
            menuItem.connect_object("activate", self.on_cipher_activate, 
                className)
            menuItem.show()
            self.cipher_menu.prepend(menuItem)
            setattr(self, className, menuItem)
        self.Gie.emit("activate")

    def on_decrypt_activate(self, w):
        """Decrypt encrypted_text_view and update decrypted_text_view."""
        self._transfer(self.encrypted_text_view, self.decrypted_text_view, 
            self._cipherGUI.decrypt)

    def on_encrypt_activate(self, w):
        """Encrypt decrypted_text_view and update encrypted_text_view."""
        self._transfer(self.decrypted_text_view, self.encrypted_text_view, 
            self._cipherGUI.encrypt)

    def on_cut_activate(self, w):
        """Do self._last_text_view.cut_clipboard."""
        self._last_text_view.emit("cut-clipboard")

    def on_copy_activate(self, w):
        """Do self._last_text_view.copy_clipboard."""
        self._last_text_view.emit("copy-clipboard")

    def on_paste_activate(self, w):
        """Do self._last_text_view.paste_clipboard."""
        self._last_text_view.emit("paste-clipboard")

    def on_clear_activate(self, w):
        """Delete the selected text in self._last_text_view, if appropriate."""
        buf = self._last_text_view.buffer
        if len(buf.selection_bounds):
            buf.delete(*buf.selection_bounds)

    def on_copy_up_activate(self, w):
        """Copy the encrypted text to the decrypted text."""
        self._transfer(self.encrypted_text_view, self.decrypted_text_view, 
            lambda s: s)

    def on_copy_down_activate(self, w):
        """Copy the decrypted text to the encrypted text."""
        self._transfer(self.decrypted_text_view, self.encrypted_text_view, 
            lambda s: s)

    def on_cipher_activate(self, cipherGUIClassName):
        """Update the currently selected ciphergui.

        The ciphergui instances are created and cached on an as needed basis.
        Update the key menu item and button so that they are only sensitive if
        the selected ciphergui has a key dialog.
        
        """
        if self._cipherGUIInstances.has_key(cipherGUIClassName):
            self._cipherGUI = self._cipherGUIInstances[cipherGUIClassName]
        else:
            CipherGUIModule = __import__("ciphergui." + cipherGUIClassName, 
                globals(), locals(), cipherGUIClassName)
            CipherGUIClass = getattr(CipherGUIModule, cipherGUIClassName)
            self._cipherGUI = self._cipherGUIInstances[cipherGUIClassName] = (
                CipherGUIClass())
        self.key_button.sensitive = self.key.sensitive = (
            hasattr(self._cipherGUI, "showKeyDialog"))

    def on_key_activate(self, w):
        """Show the key dialog for the currently selected cipher."""
        self._cipherGUI.showKeyDialog()

    def on_contents_activate(self, w):
        """Show the man page."""
        gnome.url_show("man:%s.1" % Const.NAME)

    def on_about_activate(self, w):
        """Show the about dialog."""
        gnome.ui.About(Const.NAME, Const.VERSION, Const.COPYRIGHT, Const.INFO, 
            Const.AUTHORS).show()

    def on_text_view_focus(self, text_view, e):
        """Set self._last_text_view."""
        self._last_text_view = GtkAttributesFacade(text_view)

    def _transfer(self, input_text_view, output_text_view, translator):
        """Basically output_text_view = translator(input_text_view)."""
        in_buf = input_text_view.buffer
        output_text_view.buffer.text = translator(
            in_buf.get_text(in_buf.start_iter, in_buf.end_iter, True))



"""This module contains the LibGladeApplication class."""

import signal
import gtk
from gtk import glade

from GtkAttributesFacade import GtkAttributesFacade 


class LibGladeApplication:

    """This is the base class for applications that use Glade.
    
    The following attributes are used:

    xml - This is an instance of glade.XML which encapsulates the Glade GUI.

    The following private variables are used:

    _signalsAreDone - The UNIX signal handlers only have to be set once. 

    """

    _signalsAreDone = False

    def __init__(self, gladeFile):
        """Setup the appropriate signal handlers and call setHandlers."""
        if not self._signalsAreDone:
            self._signalsAreDone = True
            signal.signal(signal.SIGINT, signal.SIG_DFL)
        self.xml = glade.XML(gladeFile)
        self.setHandlers()

    def setHandlers(self):
        """Automatically autoconnect all of the handlers.
        
        Any methods (even in subclasses) that start with "on_" will be treated
        as a handler that is automatically connected to by
        xml.signal_autoconnect.

        """
        handlers = {}
        for i in dir(self):
            if i.startswith("on_"):
                handlers[i] = getattr(self, i)
        self.xml.signal_autoconnect(handlers)

    def __getattr__(self, name):
        """If self doesn't have the attribute, check self.xml.

        If self.xml does have the attribute, wrap it in a GtkAttributesFacade 
        instance, cache it in self, and return it.
        
        """
        obj = self.xml.get_widget(name)
        if obj:
            obj = GtkAttributesFacade(obj) 
            setattr(self, name, obj)
            return obj
        raise AttributeError("%s instance has no attribute '%s'" % 
            (self.__class__.__name__, name))

    def on_quit_activate(self, *args):
        """Ignore args and call gtk.main_quit()."""
        gtk.main_quit()



"""This module contains the Proxy class.

Although the Proxy class has a simple API, the module as a whole is quite 
confusing.  Here are the logical entities involved:

proxy client - This is the client of the proxy.
proxied host - This is the host being proxied.
proxy - This is the proxy as a whole.

The proxy itself is devided into three parts:

proxy server - This is the server portion of the proxy.  It listens for 
connections from proxy clients.

proxy server backend - this is backend of the proxy server.  it handles a 
connection from a proxy client.

proxied host client - This is the portion of the proxy that connects to the 
proxied host.

Notice that proxy server backend and proxied host client are similar in that 
they each have a single connection.  In fact, I call them peers.

Consider the following usage scenario.  The proxy client wishes to connect to 
the proxied host.  The proxy client connects to the proxy server.  The proxy
server creates a proxy server backend.  The proxy server backend start the 
creation of a proxy host client.  If everything is successful, the proxy client
can send data to the proxy server backend.  The proxy server backend forwards
the data to the proxy host client.  The proxy host client forwards the data
to the proxied host.  The data flow above can be generalized to the following:
A proxy server backend or proxied host client receives data, fowards it to 
its peer, who fowards it to its associated proxied host or proxy client 
respectively.

The API is simple in that the user only has to worry about instantiating the 
Proxy class, catching socket.error exceptions, and calling asyncore.loop.
By the way, due to the way the code flows, the Proxy class is actually just a
pseudonym for the ProxyServer class.

Concerning line endings, the standard is "\r\n", but we need to be prepared to
receive "\n" as well.  Hence, look for "\n", and if there's a "\r" in front of
it, that's fine too.

"""

import asyncore
from asynchat import async_chat
import socket
from cStringIO import StringIO


class ProxyServer(asyncore.dispatcher):

    """This is the server portion of the proxy.  
    
    It listens for connections from proxy clients.  It is also the entry point
    into this module.  Thus it goes by the pseudonym Proxy.

    The following private variables are used: 

    _proxiedHost, _proxiedPort - These are host and port that are being 
    proxied.

    _encrypt, _decrypt - These are the encrypt and decrypt functions to use.
    
    """

    def __init__(self, proxiedHost, proxiedPort, listeningPort, encrypt, 
        decrypt):
        """Start listening on listeningPort."""
        LISTEN = 5
        self._proxiedHost = proxiedHost
        self._proxiedPort = proxiedPort
        self._encrypt = encrypt
        self._decrypt = decrypt
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.bind(("", listeningPort))
        self.listen(LISTEN)

    def handle_accept(self):
        """Accept a connection.
        
        Create a _ProxyServerBackend and a _ProxiedHostClient.  Let the
        _ProxiedHostClient instance take over processing since we now need to
        connect to the proxied host.

        """
        (socket, address) = self.accept()
        proxiedHostClient = _ProxiedHostClient(self, self._decrypt)
        proxyServerBackend = _ProxyServerBackend(socket, self._encrypt, 
                proxiedHostClient )
        proxiedHostClient.peer = proxyServerBackend 

Proxy = ProxyServer


class _Peer(async_chat):

    """This is the base class for _ProxyServerBackend and _ProxiedHostClient.

    It provides common behavior concerning data flow.

    The following attributes are provided:

    peer - This is the peer of this _Peer.  Remember that instances of
    _ProxyServerBackend and _ProxiedHostClient are peers of each other.
    
    The following private variables are used: 

    _translator - Apply this to the data before transfering the data to this
    _Peer's peer.

    _buf - This is a StringIO buffer for incoming data.
    
    """

    def __init__(self, socket, translator, peer):
        """Initialize the async_chat class."""
        async_chat.__init__(self, socket)
        self.set_terminator("\n")
        self.shutdown = False
        self.peer = peer
        self._translator = translator
        self._buf = StringIO()

    def collect_incoming_data(self, data):
        """Just buffer the data."""
        self._buf.write(data)

    def found_terminator(self):
        """Forward _translator(_buf) to peer and reset _buf."""
        line = self._translator(self._buf.getvalue() + "\n")
        self.peer.push(line)
        self._buf = StringIO()


class _ProxyServerBackend(_Peer):

    """This is backend of the proxy server.

    It handles a connection from a proxy client.

    It turns out that the baseclass is doing all the work.  However, I'll keep
    this class just in case I end up needing to do something here.  More 
    importantly, it's helpful to maintain the terminology.

    """

    pass


class _ProxiedHostClient(_Peer):

    """This is the portion of the proxy that connects to the proxied host."""

    def __init__(self, proxyServer, translator):
        """Start a connection to the proxied host."""
        _Peer.__init__(self, None, translator, None)
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.connect((proxyServer._proxiedHost, proxyServer._proxiedPort))

    def handle_connect(self):
        """Do nothing.  This must be defined."""
        pass



"""This file contains the Gie class."""

from CharacterCipher import CharacterCipher
from Tools import apply_caseless, offset


class Gie(CharacterCipher):

    """This is a simple cipher that maps a..z -> z..a.

    It is named after my wife Gina-Marie who showed it to me.
    
    """

    def encryptCharacter(self, c):
        """Encrypt and return c."""
        if not c.isalpha():
            return c
        return apply_caseless(lambda c: chr(ord('z') - offset(c)), c)

    def decryptCharacter(self, c):
        """Decrypt and return c."""
        return self.encryptCharacter(c)


# Do some testing.
if __name__ == '__main__':
    cipher = Gie()
    assert cipher.encrypt("aAbB zZ") == "zZyY aA"
    assert cipher.decrypt("aAbB zZ") == "zZyY aA"



"""This module contains the Vigenere class."""

from os.path import join

from KeyDialog import KeyDialog
import cipher.Vigenere
from GtkAttributesFacade import GtkAttributesFacade 
import Const


class Vigenere(KeyDialog, cipher.Vigenere.Vigenere):

    """This class permits the user to select a private key."""

    def __init__(self):
        """Init the GUI and the cipher."""
        KeyDialog.__init__(self, 
            join(Const.GLADEDIR, "ciphergui", "vigenere.glade"))
        cipher.Vigenere.Vigenere.__init__(self)

    def resetForm(self):
        """Copy the model settings to the form settings."""
        self.key_widget.text = self.key

    def applySettings(self):
        """Copy the form settings to the model settings.
        
        If the user has left the key in an invalid state (this should only be
        possible if the entry is completely empty), set the key to 
        self.defaultKey.
        
        """
        key = self.key_widget.text
        if not self.isValidKey(key):
            key = self.defaultKey
        self.key = key

    def on_key_widget_changed(self, key_widget):
        """Don't permit the user to type in invalid characters."""
        key_widget = GtkAttributesFacade(key_widget)
        key_widget.text = "".join(
            [i.lower() for i in key_widget.text if i.isalpha()])



"""This file contains global constants."""

NAME = "gcipher"
VERSION = "1.0"
COPYRIGHT = "Copyright (C) 2003 Shannon -jj Behrens.  All rights reserved."
INFO = """\
a simple encryption tool

Read Help/Contents to learn how to use gcipher
as a command line filter or as a network proxy."""
AUTHORS = ["Shannon -jj Behrens <jjinux@yahoo.com>"]
GLADEDIR = "/usr/share/gcipher/lib"



"""This module contains the GtkAttributesFacade class."""

from gobject import GObject


class GtkAttributesFacade:

    """Wrap a GTK instance to simplify the way attributes are referenced.

    Given a GTK instance i, make it possible for any functions i.get_foo() and
    i.set_foo(value) to be accessed via i.foo like a normal attribute.

    The following attributes are used:

    instance - This is the GTK instance that is being wrapped.

    """

    def __init__(self, instance):
        """Accept the instance."""
        self.__dict__["instance"] = instance

    def __setattr__(self, name, value):
        """Simplify the way attributes are referenced.

        When trying to do self.foo = something, if there is a
        self.instance.set_foo() method, use it.  Otherwise, just set the
        attribute in self.

        Return value so that chaining is possible.
        
        """
        setter = "set_" + name
        if hasattr(self.instance, setter):
            apply(getattr(self.instance, setter), [value])
        else:
            self.__dict__[name] = value
        return value

    def __getattr__(self, name):
        """Simplify the way attributes are referenced.

        Remember that this method is called after a failed lookup in self.  Try
        looking for self.instance.foo.  Next, try looking for a
        self.instance.get_foo() method, and call it if it exists.  Otherwise,
        raise an exception.

        If a value is successfully looked up, and if it is a subclass of 
        GObject, wrap it in a GtkAttributesFacade before returning it.
        
        """
        getter = "get_" + name
        if hasattr(self.instance, name):
            ret = getattr(self.instance, name)
        elif hasattr(self.instance, getter):
            ret = apply(getattr(self.instance, getter))
        else:
            raise AttributeError(
                "%s instance has no attribute '%s'" % 
                    (self.instance.__class__.__name__, name))
        if isinstance(ret, GObject):
            ret = GtkAttributesFacade(ret)
        return ret



"""This module contains the Rot class."""

from os.path import join

from KeyDialog import KeyDialog
import cipher.Rot
import Const


class Rot(KeyDialog, cipher.Rot.Rot):

    """This class permits the user to select a value for key."""

    def __init__(self):
        """Init the GUI and the cipher."""
        KeyDialog.__init__(self, 
            join(Const.GLADEDIR, "ciphergui", "rot.glade"))
        cipher.Rot.Rot.__init__(self)

    def resetForm(self):
        """Copy the model settings to the form settings."""
        self.key_widget.text = str(self.key)

    def applySettings(self):
        """Copy the form settings to the model settings."""
        self.key = int(self.key_widget.text)



"""This module contains the Vigenere class."""

from operator import add, sub

from CharacterCipher import CharacterCipher
from KeyedCipher import KeyedCipher
from Tools import LEN_ALPHABET, apply_caseless, offset, from_offset


class Vigenere(CharacterCipher, KeyedCipher):

    """This is a Rot-like cipher that uses a private key.

    Basically, each letter is rotated by an amount that is a function of both
    its position in the file as well as the private key.

    The following attributes are used:

    key - This is the private key.
    keyDescription - This is a description of what constitutes a valid key.
    defaultKey - This is the default key.

    The following private variables are used:

    _keyIndex - This is a value in the range 0..len(key) that gets incremented 
    module len(key) for every ASCII character as that character is encrypted 
    or decrypted.
    
    """

    keyDescription = "one or more lowercase letters"
    defaultKey = "a"

    def __init__(self):
        """Set key to defaultKey."""
        self.key = self.defaultKey

    def encrypt(self, s):
        """Extend the base class's encrypt method to set _keyIndex to 0."""
        self._keyIndex = 0
        return CharacterCipher.encrypt(self, s)

    def decrypt(self, s):
        """Extend the base class's decrypt method to set _keyIndex to 0."""
        self._keyIndex = 0
        return CharacterCipher.decrypt(self, s)

    def encryptCharacter(self, c):
        """Encrypt and return c."""
        return self._cryptCharacter(c, add)

    def decryptCharacter(self, c):
        """Decrypt and return c."""
        return self._cryptCharacter(c, sub)

    def _cryptCharacter(self, c, f):
        """Take care of non-letters as well as the _keyIndex."""
        if not c.isalpha():
            return c
        ret = apply_caseless(self._cryptCharacterBackend, c, f)
        self._keyIndex = (self._keyIndex + 1) % len(self.key)
        return ret

    def _cryptCharacterBackend(self, c, f):
        """Actually do the encryption or decryption."""
        inputPart = offset(c)
        keyPart = offset(self.key[self._keyIndex])
        desiredOffset = f(inputPart, keyPart) % LEN_ALPHABET
        return from_offset(desiredOffset)

    def isValidKey(self, key):
        """Is key a a valid key?  See keyDescription."""
        return key.isalpha() and key.islower()


# Do some testing.
if __name__ == '__main__':
    cipher = Vigenere()
    assert cipher.isValidKey("abz")
    assert not cipher.isValidKey("ABC")
    assert not cipher.isValidKey(" ")
    cipher.key = "bad"
    assert cipher.key == "bad"
    assert cipher.encrypt("bad") == "cag"
    assert cipher.decrypt("cag") == "bad"
    cipher.key = "xyz"
    assert cipher.encrypt("bca") == "yaz"
    assert cipher.decrypt("yaz") == "bca"



"""The variable cipherGUIList contains the list of ciphergui modules.

If an import of the module ciphergui.PluginCipherGUIList succeeds when
importing this module, the contents of
ciphergui.PluginCipherGUIList.cipherGUIList will be appended to this
cipherGUIList.

cipherGUIList is a list of tuples of the form:

    (className, guiName, labelName, description).

"""


cipherGUIList = [
    ("Gie", "Gie's Code", "_Gie's Code", "A simple cipher doable by hand"),
    ("Caesar", "Caesar's Code", "_Caesar's Code", "Julius Caesar's code"),
    ("Rot", "Rot", "_Rot", "Linear rotation"),
    ("Vigenere", "Vigenere", "_Vigenere", 
        "A version of Rot that uses a private key")
]
try:
    import ciphergui.PluginCipherGUIList
    cipherGUIList.extend(ciphergui.PluginCipherGUIList.cipherGUIList)
except ImportError:
    pass



"""This is a collection of tools for implementing ASCII based ciphers."""


LEN_ALPHABET = 26

def apply_caseless(f, c, *args, **kargs):
    """Apply the function f to the character c caselessly.

    The function f will receive a lower case version of c, but the return 
    value of this function will have the same case as the original c did.

    """
    ret = f(c.lower(), *args, **kargs)
    if c.isupper():
        ret = ret.upper()
    return ret

def offset(c):
    """Get the offset of c from 'a'."""
    return ord(c) - ord('a')

def from_offset(offset):
    """Given an offset from 'a', return the character."""
    return chr(ord('a') + offset)


# Do some testing.
if __name__ == '__main__':
    assert apply_caseless(lambda c: 'a', 'b') == 'a'
    assert apply_caseless(lambda c: 'a', 'B') == 'A'
    assert offset('a') == 0
    assert offset('z') == LEN_ALPHABET - 1
    assert from_offset(0) == 'a'
    assert from_offset(LEN_ALPHABET - 1) == 'z'



"""This module contains the MainCLI class."""

from sys import argv, stdin, stdout, stderr, exit

import Const
from cipher.KeyedCipher import InvalidKey
from AutomaticClass import AutomaticClass


class MainCLI:

    """This class is responsible for command-line parsing, etc."""

    def __init__(self):
        """Run as a GNOME application or as a command-line application."""
        if len(argv) > 1:
            self._runCommandLineApplication()
        else:
            self._runGnomeApplication()
            
    def _runGnomeApplication(self):
        """Run as a GNOME application.
        
        This method does not return until the application is finished.
        
        """
        # import pygtk
        # pygtk.require("2.0")
        # import gtk
        # import gnome
        # from MainGUI import MainGUI
        # gnome.program_init(Const.NAME, Const.VERSION)
        # MainGUI()
        # gtk.main() 

    def _runCommandLineApplication(self):
        """Run as a command-line application.

        See the man page for a description of all of these arguments.  This
        method does not return until the application is finished.
        
        """
        from getopt import getopt, GetoptError
        operations = []
        proxyMode = False
        try:
            opts, args = getopt(argv[1:], "c:C:k:p")
            for o, v in opts:
                if o.lower() == "-c":
                    operations.append(_Operation(
                        cipher = v, 
                        encrypt = (o == "-c"),
                        key = None 
                    ))
                elif o == "-k":
                    if not len(operations):
                        usage("key specified before any ciphers", 1)
                    operations[-1].key = v
                elif o == "-p":
                    proxyMode = True
                else:
                    raise AssertionError("unhandled option %s" % o)
            if proxyMode:
                self._runProxyMode(operations, args)
            else:
                self._runFilterMode(operations, args)
        except GetoptError, e:
            usage(e.msg, 1)

    def _runProxyMode(self, operations, args):
        """Handle the rest of the arguments and run in proxy mode.

        This method does not return.

        """
        from Proxy import Proxy
        import asyncore
        import socket
        if not len(args) == 3:
            usage("proxy mode requires exactly 3 arguments", 1)
        proxiedHost, proxiedPort, listeningPort = args
        try:
            Proxy(proxiedHost, 
                self._castToValidPort(proxiedPort), 
                self._castToValidPort(listeningPort),  
                self._composeOperations(operations),
                self._composeOperations(operations, True))
            asyncore.loop()
        except socket.error, e:
            stderr.write("%s: socket error: %s\n" % (Const.NAME, e.args[1]))
            exit(1)

    def _castToValidPort(self, port):
        """Cast and return port as a valid port or complain and exit."""
        MIN_PORT = 1
        MAX_PORT = 65536
        try:
            port = int(port)
            assert MIN_PORT <= port and port <= MAX_PORT
            return port
        except:
            stderr.write("%s: '%s' is not a valid port\n" % (Const.NAME, port))
            exit(1)

    def _runFilterMode(self, operations, args):
        """Handle the rest of the arguments and run in filter mode.

        This method does not return until the application is finished.

        """
        lenargs = len(args)
        input = stdin
        output = stdout
        try:
            if lenargs >= 1 and args[0] != "-":
                input = open(args[0], "r")
            if lenargs == 2 and args[1] != "-":
                output = open(args[1], "w")
            if lenargs > 2:
                usage("too many arguments", 1)
            operationsFunction = self._composeOperations(operations)
            while 1:
                line = input.readline()
                if not line:
                    break
                output.write(operationsFunction(line))
            if input != stdin:
                input.close()
            if output != stdin:
                output.close()
        except IOError, e:
            stderr.write("%s: %s: %s\n" % (Const.NAME, e.filename, e.strerror))
            exit(1)

    def _composeOperations(self, operations, invert=False):
        """Call getFunction for each operation and compose a closure.

        invert - Should the meaning of the encrypt attribute for each operation
        be reversed?

        """

        def closure(s):
            """Apply all of the operations on s, and return it."""
            for f in functions:
                s = f(s)
            return s

        functions = [i.getFunction(invert) for i in operations]
        return closure


class _Operation(AutomaticClass):

    """This is a simple class representing an encryption operation.
    
    The following attributes are used:

    cipher - This string is the name of a cipher.
    encrypt - If true, do encrypt.  Otherwise, do decrypt.
    key - This string is the key to use with the cipher.

    """

    def getAttributes(self):
        return AutomaticClass.getAttributes(self) + [
            "cipher", "encrypt", "key"]

    def getFunction(self, invert = False):
        """This is a factory for a function representing this _Operation.
        
        Instantiate the cipher, set the key, and return either the encrypt or
        decrypt function.  This function may call usage and exit.

        invert - Should the meaning of the encrypt attribute be reversed?

        """
        try: 
            CipherModule = __import__("cipher." + self.cipher, globals(), 
                locals(), self.cipher)
            CipherClass = getattr(CipherModule, self.cipher)
            cipherInstance = CipherClass()
            if hasattr(cipherInstance, "key"):
                if self.key != None:
                    cipherInstance.key = self.key
            else:
                if self.key != None:
                    usage("no key required for cipher '%s'" % self.cipher, 1)
            if self.encrypt or invert:
                return cipherInstance.encrypt
            else:
                return cipherInstance.decrypt
        except ImportError:
            usage("failed to import cipher '%s'" % self.cipher, 1)
        except InvalidKey, e:
            # usage("invalid key '%s' for cipher '%s'; it's not '%s'." %
            #     (e.key, operation.cipher, e.keyDescription), 1)
            usage("invalid key '%s' for cipher '%s'; it's not '%s'." %
                (e.key, self.cipher, e.keyDescription), 1)

            

def usage(warning=None, exitCode=None):
    """Print usage information.

    warning - If this is not None, it will be printed as well.

    exitCode - If this is not None, the application will exit with this 
    exit code.

    """
    if warning:
        stderr.write("%s: %s\n" % (Const.NAME, warning))
    stderr.write("Try `man %s' for more information.\n" % Const.NAME)
    if exitCode != None:
        exit(exitCode)



"""This file contains the AutomaticClass class."""


class AutomaticClass:

    """This is a base class for simple, struct-like classes."""

    def __init__(self, **kargs):
        """Subclasses should call this after initializing any defaults.

        **kargs - Set each attribute with the given value.  Keys in kargs that
        aren't in self.__attributes__ will be ignored.  

        """
        for i in self.__attributes__:
            if kargs.has_key(i):
                setattr(self, i, kargs[i])

    def getAttributes(self):
        """Return the list of attributes that this class uses.
        
        Subclasses should extend (not override) this method to add their own
        attributes.
        
        """ 
        return []

    def __getattr__(self, attribute):
        """Handle the following attributes manually:  __attributes__.
        
        Meta, meta, meta!
        
        """
        if attribute == "__attributes__":
            return self.getAttributes()
        raise AttributeError

    def __repr__(self):
        """Return a suitable string representing of this instance."""
        attributes = []
        for i in self.__attributes__:
            try: 
                value = getattr(self, i)
            except: 
                value = "Undefined"
            attributes.append("%s=%s" % (i, value))
        return "<%s %s>" % (self.__class__.__name__, " ".join(attributes))


# This is for testing.
class _Test(AutomaticClass):
    def __init__(self, **kargs):
        self.foo = "Bar"
        AutomaticClass.__init__(self, **kargs)
    def getAttributes(self):
        return AutomaticClass.getAttributes(self) + ["foo"]


# Do some testing.
if __name__ == '__main__': 
    assert `_Test()` == "<_Test foo=Bar>"
    assert _Test().foo == "Bar"
    assert _Test(foo = "Not bar").foo == "Not bar"



"""This module contains the KeyedCipher class and the InvalidKey exception."""


class KeyedCipher:

    """This is a simple base class for ciphers with keys.

    Subclasses of this class should do the following:

    1) Define and initialize an attribute named key.

    2) Define an attribute named keyDescription which holds a sentence 
       fragment describing what constitutes a valid key.

    3) Defined a method isValidKey(self, key) that returns True if the given
       key is valid.

    4) Possibly define a defaultKey attribute.

    5) Extend __setattr__ instead of overriding it.

    """

    def __setattr__(self, attribute, value):
        """Perform validation for the key attribute."""
        if attribute == "key":
            if not self.isValidKey(value):
                self.raiseInvalidKey(value)
        self.__dict__[attribute] = value

    def raiseInvalidKey(self, key):
        """Raise an InvalidKey exception for the given key."""
        raise InvalidKey(key, self.keyDescription)


class InvalidKey(ValueError):

    """An attempt was made to set an invalid key.
    
    The following attributes are used:

    key - This is the faulty key.

    keyDescription - This is a sentence fragment describing what is expected
    for the key.
    
    """

    def __init__(self, key, keyDescription):
        """Accept the arguments."""
        self.key = key
        self.keyDescription = keyDescription
        ValueError.__init__(self, key, keyDescription)



"""This module contains the KeyDialog class."""

from LibGladeApplication import LibGladeApplication


class KeyDialog(LibGladeApplication):

    """This is a baseclass for ciphergui's that have key dialogs.

    The subclass should override the resetForm and applySettings methods.

    """

    def __init__(self, gladeFile):
        """Call LibGladeApplication's constructor."""
        LibGladeApplication.__init__(self, gladeFile)

    def showKeyDialog(self):
        """Show the key dialog for this cipher."""
        self.dialog.show()

    def on_dialog_show(self, w):
        """Call self.resetForm()."""
        self.resetForm()

    def on_dialog_delete_event(self, w, e):
        """Hide the dialog and return True."""
        self.dialog.hide()
        return True

    def on_cancelbutton_clicked(self, w):
        """Hide the dialog."""
        self.dialog.hide()

    def on_applybutton_clicked(self, w):
        """Call self.applySettings()."""
        self.applySettings()

    def on_okbutton_clicked(self, w):
        """Call self.applySettings() and hide the dialog."""
        self.applySettings()
        self.dialog.hide()

    def resetForm(self):
        """Copy the model settings to the form settings."""
        pass

    def applySettings(self):
        """Copy the form settings to the model settings."""
        pass



"""This module contains the Rot class."""

from CharacterCipher import CharacterCipher
from KeyedCipher import KeyedCipher
from Tools import LEN_ALPHABET, apply_caseless, offset, from_offset


class Rot(CharacterCipher, KeyedCipher):

    """This is the cipher used in Rot7, Rot13, Caesar's code, etc.

    This is a simple cipher that simply rotates the characters by an offset.
    For instance, ab..z -> b..za.

    The following attributes are used:

    key - What is the offset used in the rotation?  This will initially be set
    to defaultKey.

    keyDescription - This is a description of what constitutes a valid key.

    defaultKey - This is the default key.
    
    """

    keyDescription = "an integer in the range [0, 25]"
    defaultKey = 0

    def __init__(self):
        """Set key to defaultKey."""
        self.key = self.defaultKey

    def encryptCharacter(self, c):
        """Encrypt and return c."""
        return self._encryptCharacter(c, self.key)

    def decryptCharacter(self, c):
        """Decrypt and return c."""
        return self._encryptCharacter(c, -self.key)

    def _encryptCharacter(self, c, key):
        """Encrypt and return c.  Use the given key."""
        if not c.isalpha():
            return c
        return apply_caseless(
            lambda c: from_offset((offset(c) + key) % LEN_ALPHABET), c)

    def isValidKey(self, key):
        """Is key a a valid key?  See keyDescription."""
        return isinstance(key, type(0)) and 0 <= key and key < LEN_ALPHABET

    def __setattr__(self, attribute, value):
        """Extend KeyedCipher.__setattr__ to cast the key to an int."""
        strvalue = value
        if attribute == "key":
            try:
                value = int(value)
            except ValueError:
                self.raiseInvalidKey(strvalue)
        KeyedCipher.__setattr__(self, attribute, value)


# Do some testing.
if __name__ == '__main__':
    cipher = Rot()
    assert cipher.isValidKey(0)
    assert not cipher.isValidKey(-1)
    assert not cipher.isValidKey("a")
    cipher.key = "1"
    assert cipher.key == 1
    assert cipher.encrypt("aAzZ") == "bBaA"
    assert cipher.decrypt("bBaA") == "aAzZ"
